
-- # import Property
-- # import Pairs

{------------------------------------------------------------------------
    Formalization of a category
 ------------------------------------------------------------------------}

Cat : #2
Cat = sig Obj	  : Type
	  Arr	  : (a, b : Obj) -> Set   -- not good if all sets are inhabited!
	  id	  : (a : Obj) -> Arr a a
	  comp	  : (a, b, c : Obj) -> Arr b c -> Arr a b -> Arr a c
	  axId1	  : (a, b : Obj) -> (f : Arr a b) -> comp a b b (id b) f == f
	  axId2	  : (a, b : Obj) -> (f : Arr a b) -> comp a a b f (id a) == f
	  axComp  : (a, b, c, d : Obj) ->
		     (f : Arr c d) -> (g : Arr b c) -> (h : Arr a b) ->
		     comp a c d f (comp a b c g h) == comp a b d (comp b c d f g) h

Obj : (C : Cat) -> Type
Obj C = C.Obj

Arr (C : Cat) : Obj C -> Obj C -> Set
Arr a b = C.Arr a b

infixr 8 -->
(-->) (C : Cat) : Obj C -> Obj C -> Set
a --> b = Arr a b

id (C : Cat)(a : Obj C) : a --> a
id = C.id a

infixr 10 @
(@) (C : Cat)(a, b, c : Obj C) : (b --> c) -> (a --> b) -> (a --> c)
f @ g = C.comp a b c f g

axIdLeft (C : Cat)(a, b : Obj C) : (f : a --> b) -> id @ f == f
axIdLeft f = C.axId1 a b f

axIdRight (C : Cat)(a, b : Obj C) : (f : a --> b) -> f @ id == f
axIdRight f = C.axId2 a b f

axId (C : Cat)(a, b : Obj C) : (f : a --> b) -> id @ f == f /\ f @ id == f
axId f = fol-plugin(axIdLeft f, axIdRight f)

axComp (C : Cat)(a, b, c, d : Obj C) :
	(f : c --> d) -> (g : b --> c) -> (h : a --> b) ->
	f @ (g @ h) == (f @ g) @ h
axComp f g h = C.axComp a b c d f g h

{------------------------------------------------------------------------
    The dual of a category
 ------------------------------------------------------------------------}

op : Cat -> Cat
op C =
  struct  Obj	  = Obj C
	  Arr	  = \a b -> Arr {C} b a
	  id	  = \a -> id
	  comp	  = \a b c f g -> g @ f
	  axId1	  = \a b f -> axIdRight f
	  axId2	  = \a b f -> axIdLeft f
	  axComp  = \a b c d f g h -> fol-plugin(axComp h g f)

-- opObj (C : Cat) : Obj C -> Obj (opCat C)
-- opObj a = a
-- 
-- opArr (C : Cat)(a, b : Obj C) : Arr {C} a b -> Arr {opCat C} (opObj b) (opObj a)
-- opArr g = g

{------------------------------------------------------------------------
    The product category
 ------------------------------------------------------------------------}

(<*>) : Cat -> Cat -> Cat
C <*> D =
  struct  Obj	  = { Obj C ; Obj D }
	  Arr	  = \a b -> { a.1 --> b.1 ; a.2 --> b.2 }
	  id	  = \a -> (id, id)
	  comp	  = \a b c f g -> (f.1 @ g.1, f.2 @ g.2)
	  axId1	  = \a b f -> fol-plugin( freeze (@) id
					, axIdLeft
					, surjectivePairing
					)
	  axId2	  = \a b f -> fol-plugin( freeze (@) id
					, axIdRight
					, surjectivePairing
					)
	  axComp  = \a b c d f g h -> fol-plugin( freeze (@) id, axComp )

{------------------------------------------------------------------------
    The category of arrows of a category
    (We can't define this, since there are side condition on the arrows)
 ------------------------------------------------------------------------}

-- arrowCat : Cat -> Cat
-- arrowCat C =
--   struct  Obj    = sig dom   : Obj C
--                        codom : Obj C
--                        arr   : Arr dom codom
--           Arr     = \a b -> 

{------------------------------------------------------------------------
    Mono and epi
 ------------------------------------------------------------------------}

isMono (C : Cat)(b, c : Obj C) : (f : b --> c) -> Type
isMono f = (a : Obj C) -> (g, h : a --> b) -> f @ g == f @ h ==> g == h

isEpi (C : Cat)(b, c : Obj C) : (f : c --> b) -> Type
isEpi f = isMono {op C} f

monoEpiDuality (C : Cat)(b, c : Obj C) :
      (f : b --> c) -> isMono f -> isEpi {op C} f
monoEpiDuality f monf = \a -> fol-plugin(monf a)

monoComp (C : Cat)(a, b, c : Obj C) :
    (f : b --> c) -> (g : a --> b) ->
    isMono (f @ g) -> isMono g
monoComp f g mon_fg d = fol-plugin(freeze (@), mon_fg d, axComp)

{------------------------------------------------------------------------
    Initial and terminal objects
 ------------------------------------------------------------------------}

isIsomorphism (C : Cat)(a, b : Obj C) : (a --> b) -> (b --> a) -> Prop
isIsomorphism i j = i @ j == id /\ j @ i == id

Isomorphism (C : Cat) : (a, b : Obj C) -> Type
Isomorphism a b = sig i	  : a --> b
		      j	  : b --> a
		      iso : i @ j == id /\ j @ i == id

isUnique (C : Cat) : (Obj C -> Type) -> Obj C -> Type
isUnique P a = (b : Obj C) -> P b -> Isomorphism a b

uniqueArrow (C : Cat) (a, b : Obj C) : ((a --> b) -> Prop) -> Type
uniqueArrow P = sig h	  : Arr a b
		    uniq  : (h' : a --> b) -> P h' ==> h == h'

isUniversal (C : Cat) : (Obj C -> Type) -> Type
isUniversal P = (a : Obj C) -> P a -> isUnique P a

isTerminal (C : Cat) : Obj C -> Type
isTerminal one = (a : Obj C) -> uniqueArrow {C} {a} {one} (\h -> True)

terminalUnique (C : Cat) : isUniversal {C} isTerminal
terminalUnique a terminal_a b terminal_b =
    struct  i	= (terminal_b a).h
	    j	= (terminal_a b).h
	    iso	= fol-plugin(freeze (@) id, (terminal_b a).uniq, (terminal_a b).uniq)
		    -- bad reasoning(?)

