-- Dining Philosophers
-- Author:: Thierry Coquand

-- # import Property

-- Lists ------------------------------------------------------------------

List : Set -> Set
nil : (A:Set) -> List A
cons : (A:Set) -> A -> List A -> List A

listInd : (A	    :	Set) ->
	   (P	    :	List A -> Prop) ->
	   (P_nil  :	P (nil A)) ->
	   (P_cons :	(x:A) -> (xs:List A) ->
		P xs -> P (cons A x xs)) ->
	   (xs	    :	List A) ->
	   P xs


------------ the set of philosophers -------------------

P : Set
S : P -> P -> Prop
symS : (x:P) -> (y:P) -> S x y ==> S y x

------------- the set of events -------------------------

Ev : Set
T : P -> Ev
E : P -> Ev
L : P -> Ev

eventInd : (Q : Ev -> Prop) ->
            ((a:P) -> Q (T a)) ->
            ((a:P) -> Q (E a)) ->
            ((a:P) -> Q (L a)) ->
            (e:Ev) -> Q e

------------- the set of event traces ----------------------

Evt : Set
Evt = List Ev

nl : List Ev
nl = nil Ev

infixr 7 ::
(::) : Ev -> Evt -> Evt
(::) = cons Ev

--- valid event traces ------------

pos : Evt -> Ev -> Prop
val : Evt -> Prop

axValN : val nl
axValC : (es:Evt)->(e:Ev) -> ( val (e::es) <=>  val es /\ pos es e )

---- the observable predicates and relations ----------

R : Evt -> P -> P -> Prop
tab : Evt -> P -> Prop
eat : Evt -> P -> Prop

isFirst : Evt -> P -> Prop

---- the protocole ---------------

SYS : (es:Evt) -> (a:P) -> (x:P) -> (y:P) ->
       (  
           (R ((T a)::es) x y   <=>  (R es x y \/ ((Eq P a x) /\ tab es y /\ S a y)))
        /\ (R ((E a)::es) x y   <=>  R es x y)
        /\ (R ((L a)::es) x y   <=> (R es x y /\ Not (Eq P a y)))
        /\ (eat ((E a)::es) x   <=> (eat es x \/ Eq P a x))
        /\ (eat ((T a)::es) x   <=> eat es x)
        /\ (eat ((L a)::es) x   <=> (eat es x /\ Not (Eq P a x)))
        /\ (tab ((E a)::es) x   <=> tab es x)
        /\ (tab ((T a)::es) x   <=> tab es x \/ Eq P a x)
        /\ (tab ((L a)::es) x   <=> (tab es x /\ Not (Eq P a x)))
        /\ (pos es (T a)       <=> Not (tab es a))
        /\ (pos es (E a)       <=> (tab es a /\ isFirst es a))
        /\ (pos es (L a)       <=> eat es a)
        /\ (isFirst es a        ==> Not (R es a x))
        /\ Not (R nl x y)
        /\ Not (eat nl x)
        /\ Not (tab nl x)
       )


--- some invariants -------

---- invariant 0 -----------

prop : (es:Evt) -> Prop
prop es = (x:P) -> (y:P) -> val es ==> eat es x ==> tab es x

lemT : (es:Evt) -> prop es  -> (a:P) -> prop ((T a)::es)
lemT = \ es -> \ h1 -> \ a -> fol-plugin (h1,SYS,axValC)   

lemE : (es:Evt) -> prop es  -> (a:P) -> prop ((E a)::es)
lemE = \ es -> \ h1 -> \ a -> 
             fol-plugin (h1,SYS,axValC)

lemL : (es:Evt) -> prop es  -> (a:P) -> prop ((L a)::es)
lemL = \ es -> \ h1 -> \ a -> fol-plugin (h1,SYS,axValC)

lem : (es:Evt) -> prop es -> (e:Ev) -> prop (e::es)
lem = \ es -> \ h -> eventInd (\ e -> prop (e::es)) (lemT es h) (lemE es h) (lemL es h)

thm : (es:Evt) -> prop es
thm = listInd Ev prop fol-plugin(SYS) (\ e -> \ es -> \ h -> lem es h e)

---- invariant 1 -----------

prop1 : (es:Evt) -> Prop
prop1 es = (x:P) -> (y:P) -> val es ==> (   (R es x y ==> (tab es x /\ tab es y))
                                           /\ ((eat es x /\ tab es y) ==> Not (R es x y)) )

lem1T : (es:Evt) -> prop1 es -> (a:P) ->  prop1 ((T a)::es)
lem1T = \ es -> \ h1 -> \ a -> fol-plugin (h1,SYS,axValC,thm)

lem1E : (es:Evt) -> prop1 es -> (a:P)  -> prop1 ((E a)::es)
lem1E = \ es -> \ h1 -> \ a -> fol-plugin (h1,SYS,axValC)

lem1L : (es:Evt) -> prop1 es -> (a:P)  -> prop1 ((L a)::es)
lem1L = \ es -> \ h1 -> \ a -> fol-plugin (h1,SYS,axValC)

lem1 : (es:Evt) -> prop1 es -> (e:Ev) -> prop1 (e::es)
lem1 = \ es -> \ h -> eventInd (\ e -> prop1 (e::es)) (lem1T es h) (lem1E es h) (lem1L es h)
      
thm1 : (es:Evt) -> prop1 es
thm1 = listInd Ev prop1 fol-plugin(SYS) (\ e -> \ es -> \ h -> lem1 es h e)

--- absence of deadlock -------

acyclic : (P->P->Prop) -> Prop

sub : (P->P->Prop) -> (P->P->Prop) -> Prop
sub = \ R1 -> \ R2 -> (x:P) -> (y:P) -> R1 x y -> R2 x y

lem1Sub : (R1:P->P->Prop) -> (R2:P->P->Prop) -> sub R1 R2 -> acyclic R2 -> acyclic R1

lem2Sub : (R1:P->P->Prop) -> (R2:P->P->Prop) -> (a:P) -> ((x:P) -> Not (R1 a x)) ->
           ((x:P) -> (y:P) -> R2 x y -> (R1 x y \/ Eq P x a)) ->
           acyclic R1 -> acyclic R2
        
Main : (es:Evt) -> Prop
Main es = acyclic (R es)

lemMT : (es:Evt) -> (a:P) -> Main es  -> val ((T a)::es) -> Main ((T a)::es)
lemMT = \ es -> \ a -> \ h1 -> \ h2 ->
           let
             rem1 : (x:P) -> Not (R es a x) 
             rem1 = \ x -> fol-plugin (SYS,axValC,h2,thm1)
             rem2 : (x:P) -> (y:P) -> R ((T a)::es) x y ==> (R es x y \/ Eq P x a)
             rem2 = \ x -> \ y -> fol-plugin (SYS,h2,axValC)
           in lem2Sub (R es) (R ((T a)::es)) a rem1 rem2 h1

lemME : (es:Evt) -> (a:P) -> Main es  -> Main ((E a)::es)
lemME = \ es -> \ a -> \ h1 -> 
           let
             rem : sub (R ((E a)::es)) (R es) 
             rem = \ x -> \ y -> \ h2 -> fol-plugin (h2,SYS,axValC)
           in lem1Sub (R ((E a)::es)) (R es) rem h1


lemML : (es:Evt) -> (a:P) -> Main es  -> Main ((L a)::es)
lemML = \ es -> \ a -> \ h1 -> 
           let
             rem : sub (R ((L a)::es)) (R es) 
             rem = \ x -> \ y -> \ h2 -> fol-plugin (h2,SYS,axValC)
           in lem1Sub (R ((L a)::es)) (R es) rem h1

main : (es:Evt) -> val es -> Main es
