
-- # import Rel
-- # import Setoid
-- # import ChainReasoning

sig Preorder (A:SET) : Type
  extends ReflexiveRel
	, TransitiveRel where

sig PartialOrder (A:Setoid) : Type
  extends Preorder
	, AntiSymmetricRel
	, ReflexiveSetoidRel where

sig TotalOrder (A:Setoid) : Type
    extends PartialOrder, TotalRel where

sig StrictPartialOrder (A:Setoid) : Type
  extends IrreflexiveRel
	, AsymmetricRel
	, TransitiveRel
	, SetoidRel where

sig StrictTotalOrder (A:Setoid) : Type
    extends StrictPartialOrder where
  total : (x, y : El A) -> Prf (rel x y \/ x == y \/ rel y x)

strictPartialOrder (A:Setoid) : PartialOrder A -> StrictPartialOrder A
strictPartialOrder {A} po =
  struct rel	= (<)
	 irrefl = irrefl
	 asym	= asym
	 trans	= trans
	 respeq = respeq
  where
    infix 10 <=, <

    (<=) = po.rel

    (<) : El A -> El A -> PROP
    x < y = x <= y /\ Not (x == y)

    less = mkRel (<)

    respeq : Prf (RespectsEq less)
    respeq x y z w "x=y" "z=w" (andI "x<=z" "x!=z") =
      andI (po.respeq _ _ _ _ "x=y" "z=w" "x<=z") "y!=w"
      where
	"y!=w" : Prf (Not (y == w))
	"y!=w" "y=w" =
	  notE "x!=z"
	  $ chainReasoning
	  $ x ===| y  `since` "x=y"
	      |==| w  `since` "y=w"
	      |==| z  `since` symmetry "z=w"

    irrefl : Prf (Irreflexive less)
    irrefl x y eqxy (andI _ neqxy) = notE neqxy eqxy

    asym : Prf (Asymmetric less)
    asym x y (andI leqxy neqxy) (andI leqyx neqyx) =
      neqxy (po.antisym _ _ leqxy leqyx)

    trans : Prf (Transitive less)
    trans x y z (andI "x <= y" "x != y")
		(andI "y <= z" "y != z")
		= andI "x <= z" "x != z"
      where
	"x <= z" = po.trans _ _ _ "x <= y" "y <= z"
	"x != z" : Prf (Not (x == z))
	"x != z" eqxz = "x != y" $ po.antisym _ _ "x <= y" "y <= x"
	  where
	    "z <= x" = setoidRefl po _ _ (A.eq.sym _ _ eqxz)
	    "y <= x" = po.trans _ _ _ "y <= z" "z <= x"

nonStrictPartialOrder (A:Setoid) : StrictPartialOrder A -> PartialOrder A
nonStrictPartialOrder {A} spo =
  struct rel	 = (<=)
	 refl	 = refl
	 trans	 = trans
	 antisym = antisym
	 respeq  = respeq
  where
    infix 10 <, <=

    (<) = spo.rel
    
    (<=) : El A -> El A -> PROP
    x <= y = x < y  \/  x == y

    leq = mkRel (<=)

    refl : Prf (Reflexive leq)
    refl x = orI2 (A.eq.refl x)

    respeq : Prf (RespectsEq leq)
    respeq x y z w "x=y" "z=w" (orI1 "x<z") =
      orI1 $ spo.respeq x y z w "x=y" "z=w" "x<z"
    respeq x y z w "x=y" "z=w" (orI2 "x=z") =
      orI2 $ chainReasoning
	   $ y	===| x	`since` symmetry "x=y"
		|==| z	`since` "x=z"
		|==| w	`since` "z=w"

    antisym : Prf (AntiSymmetric leq)
    antisym x y (orI2 eqxy)  _		= eqxy
    antisym x y (orI1 _)    (orI2 eqyx) = A.eq.sym _ _ eqyx
    antisym x y (orI1 ltxy) (orI1 ltyx) = falseE $ spo.asym _ _ ltxy ltyx

    trans : Prf (Transitive leq)
    trans x y z (orI1 ltxy) (orI1 ltyz) =
      orI1 $ spo.trans _ _ _ ltxy ltyz
    trans x y z (orI1 ltxy) (orI2 eqyz) =
      orI1 $ spo.respeq _ _ _ _ (A.eq.refl _) eqyz ltxy
    trans x y z (orI2 eqxy) (orI1 ltyz) =
      orI1 $ spo.respeq _ _ _ _ (A.eq.sym _ _ eqxy) (A.eq.refl _) ltyz
    trans x y z (orI2 eqxy) (orI2 eqyz) =
      orI2 $ A.eq.trans _ _ _ eqxy eqyz

