
module Plugins.Assert where

import Text.PrettyPrint
import Control.Monad	    ( when )
import Control.Monad.Trans  ( liftIO )

import Interface.Options    ( Options(..) )

import Syntax.Internal	    ( AnnotatedTerm(..), Term, Value(..) )
import Syntax.Pretty	    ( Pretty(..) )

import TypeChecker.Monad
import TypeChecker.Monad.Errors
import TypeChecker.Monad.Options
import TypeChecker.UnfoldConstants

import Utils.Error

import Plugins

type AssertOptions = ()
type AssertState   = [Assertion]

data Assertion = Assertion Value

instance Pretty Assertion where
    pretty (Assertion v) = 
        sep [ text "  ? ::" <+> pretty v
	    , text "    ==" <+> pretty (unfoldAllConstants v)
            ]

instance Show Assertion where
    show = show . pretty

assertCheck :: AssertOptions -> AssertState -> [Term] -> Value ->
               TC (Value, AssertState)
assertCheck o s []    v		    = return (VTop v, s ++ [Assertion v])
assertCheck o s [Var "silent" _] v  = return (VTop v, s)
assertCheck o s [Def "silent" _] v  = return (VTop v, s)
assertCheck o s (_:_) v		    = failure assertWithArgumentsError 

assertFinal :: AssertOptions -> AssertState -> TC ()
assertFinal o [] = return ()
assertFinal o s = 
    do opts <- commandLineOptions 
       when (optVerbose opts >= 1) $
            liftIO $ do print $ hcat $ [ text "=== Assertions ",
                                         text $ replicate 55 '=' ]
                        mapM_ print s 

assertHelp :: String
assertHelp =
    unlines [ "Asserts existence of object."
	    , "  assert-plugin()       -- prints the asserted type if verbosity > 0"
	    , "  assert-plugin(silent) -- doesn't print asserted type"
	    ]

plugin :: SomePlugin
plugin = mkPlugin
    Plugin  { plName		= "assert"
            , plHelp            = assertHelp
	    , plInit		= \_ -> return []
	    , plFinal		= assertFinal
	    , plCheck		= assertCheck
	    , plInfer		= \_ _ _ -> failure noInferenceError
	    , plOptions		= []
	    , plDefaultOptions	= ()
	    }

noInferenceError :: ErrorFun
noInferenceError e =
    e	{ errWhat   = fsep $ pwords "type inference failed"
	, errWhy    = fsep $ pwords "the type of an assertion cannot be inferred"
	}

assertWithArgumentsError :: ErrorFun
assertWithArgumentsError e =
    e { errWhat = fsep $ pwords "assertion failed"
      , errWhy  = fsep $ pwords "assert does not accept arguments"
      }
