
module TypeChecker.Monad.Definitions
    ( getDefinitions
    , lookupDef
    , addDefinition
    , addDefinitions
    , setDefinitions
    )
    where

import Control.Monad.State

import Syntax.Internal
import TypeChecker.Monad
import TypeChecker.Monad.Errors
import TypeChecker.Errors
import Utils.Map

-- Definitions ------------------------------------------------------------

getDefinitions :: TC Definitions
getDefinitions = liftM tcDefinitions get

-- Note: don't export
updateDefinitions f s = s { tcDefinitions = f $ tcDefinitions s }

lookupDef :: Name -> TC DefValue
lookupDef x =
    do	delta <- getDefinitions
	case lookupMap x delta of
	    Just v  -> return v
	    _	    -> failure $ notDefinedInternalError x

addDefinition :: Name -> DefValue -> TC ()
addDefinition x d = modify (updateDefinitions $ updateMap x d)

addDefinitions :: [(Name, DefValue)] -> TC ()
addDefinitions ds = mapM_ (uncurry addDefinition) ds

setDefinitions :: Definitions -> TC ()
setDefinitions defs = modify (updateDefinitions $ const defs)

