
module TypeChecker.Patterns.Match where

import Control.Monad
import Data.Monoid
import Syntax.Internal
import Syntax.Values

import TypeChecker.Values
import TypeChecker.Monad
import {-# SOURCE #-} TypeChecker.UnfoldConstants -- REMOVE_IF_HADDOCK
import Utils.Monads

{------------------------------------------------------------------------
    Pattern matching
 ------------------------------------------------------------------------}

data MatchResult
	= Match [Value]
	| DontKnow
	| MisMatch

instance Monoid MatchResult where
    mempty			= Match []

    MisMatch `mappend` _	= MisMatch
    _ `mappend` MisMatch	= MisMatch
    DontKnow `mappend` _	= DontKnow
    _ `mappend` DontKnow	= DontKnow
    Match vs `mappend` Match ws = Match (vs ++ ws)

-- | Match a pattern against a value.
patternMatch :: Pattern -> Value -> TC (Value, MatchResult)
patternMatch (WildP _) v	= return (v, Match [])
patternMatch (VarP _) v		= return (v, Match [v])
patternMatch (PairP _ p1 p2) v  =
    do	(v1, m1) <- patternMatch p1 (vProject v VFst)
	(v2, m2) <- patternMatch p2 (vProject v VSnd)
	return (vPair v1 v2, mappend m1 m2)
patternMatch (ConP _ x hs ps) v =
    do	v' <- unfoldConstants v
	case vAppView v' of
	    VAVCon y h ar vs | x == y  ->
		do  (vs',ms) <- unzip <$> zipWithM patternMatch (hs ++ ps) vs
		    return (vApps (vCon y h ar) vs', mconcat ms)
	    VAVDef y _ _    -> return (v', DontKnow)
	    _		    -> return (v', MisMatch)

